__all__ = ["P", "E", "Var", "Std"]

import functools
from typing import Any, Callable, Generic, Tuple, TypeVar, Union

import mpmath

from hall._core import Stochast
from hall._types import C, Complex
from hall._types import Float
from hall._types import Float as Probability
from hall.event import Event


_F = TypeVar("_F", bound=Callable[..., Complex])


class Operator(Generic[_F]):
    __wrapped__: _F

    def __init__(self, __func: _F, /):
        self.__wrapped__ = __func
        functools.update_wrapper(self, self.__wrapped__)

    def __call__(self, *__vals) -> Complex:
        return self.__wrapped__(*__vals)

    def __getitem__(self, __vals: Union[Tuple[Any, ...], Any]) -> Complex:
        if isinstance(__vals, tuple):
            return self.__wrapped__(*__vals)
        return self.__wrapped__(__vals)


@Operator
def P(event: Event[C]) -> Probability:
    """
    Probability measure for the given event, i.e. the probability that the
    event occurs
    """
    return event.p


@Operator
def E(X: Stochast[C]) -> Union[C, Float]:
    """Expected value"""
    return X.distribution.mean


@Operator
def Var(X: Stochast[C]) -> Union[C, Float]:
    """Variance"""
    return X.distribution.variance


@Operator
def Std(X: Stochast[C]) -> Union[C, Float]:
    """Standard deviation"""
    return mpmath.mp.sqrt(X.distribution.variance)


# TODO Covariance (Cov)
#  https://en.wikipedia.org/wiki/Covariance

# TODO Correlation
#  https://en.wikipedia.org/wiki/Pearson_correlation_coefficient



__all__ = [
    "Bernoulli",
    "Uniform",
    "Binomial",
    "Bern",
    "U",
    "B",
]

import abc
import numbers
from typing import ClassVar, Final, Protocol, Union

import mpmath

from hall._core import Distribution as _Distribution
from hall._types import Float
from hall._types import Float as Probability
from hall._types import Integral, Interval, Z, is_probability


class Distribution(_Distribution[Z], Protocol[Z]):
    discrete: ClassVar[bool] = True

    @abc.abstractmethod
    def pmf(self, x: Z) -> Probability:
        """Probability Mass Function"""
        ...

    @abc.abstractmethod
    def cmf(self, x: Z) -> Probability:
        """Cumulative Mass Function"""
        ...

    def cmf_inv(self, y: Probability) -> Z:
        """Inverse of the cmf"""
        support = self.support
        if not support.is_bounded:
            raise NotImplementedError

        x_prev = 0
        for x in range(support.a, support.b + 1):
            # noinspection PyTypeChecker
            yi = self.cmf(x)
            assert mpmath.mp.zero <= yi <= mpmath.mp.one

            if yi > y:
                break

            x_prev = x

        # noinspection PyTypeChecker
        return x_prev

    def f(self, x: Z) -> Probability:
        if not isinstance(x, numbers.Integral):
            raise TypeError("value must be an integral number")

        if x not in self.support:
            return mpmath.mpf(0)

        return self.pmf(x)

    def F(self, x: Z) -> Probability:
        if not isinstance(x, numbers.Integral):
            raise TypeError("value must be an integral number")

        support = self.support
        if x < support:
            return mpmath.mpf(0)
        if x > support:
            return mpmath.mpf(1)

        return self.cmf(x)

    def G(self, y: Probability) -> Z:
        if not (0 <= y <= 1):
            raise TypeError("probability must be between 0 and 1 inclusively")

        return self.cmf_inv(y)


class Binomial(Distribution[Z]):
    __slots__ = ("n", "p")

    n: Z
    p: Final[Probability]

    def __init__(self, n: Z, p: Probability):
        if not isinstance(n, numbers.Integral):
            raise TypeError("n must be an integral number")
        if n < 0:
            raise ValueError("n must be postive")
        if not is_probability(p):
            raise ValueError("p must lie between 0 and 1 inclusively")

        self.n = n
        self.p = mpmath.mpf(p)

    @property
    def support(self) -> Interval[Z]:
        return Interval(type(self.n)(0), self.n)

    @property
    def q(self) -> Probability:
        return mpmath.mpf(1) - self.p

    @property
    def mean(self) -> Float:
        return self.n * self.p

    @property
    def variance(self) -> Float:
        return self.n * self.p * self.q

    def pmf(self, x: Integral) -> Probability:
        if x < self.support:
            raise TypeError(f"x must be larger than {self.support.a}")
        if self.support < x:
            return mpmath.mpf(0)

        return mpmath.binomial(self.n, x) * self.p ** x * self.q ** (self.n - x)

    def cmf(self, x: Integral) -> Probability:
        if x < self.support:
            raise TypeError(f"x must be larger than {self.support.a}")
        if self.support < x:
            return mpmath.mpf(1)

        return mpmath.betainc(self.n - x, x + 1, 0, self.q, regularized=True)


class Bernoulli(Binomial):
    __slots__ = ()

    def __init__(self, p: Probability):
        # noinspection PyTypeChecker
        super().__init__(1, p)

    def cmf_inv(self, y: Probability) -> int:
        if y <= mpmath.mp.one - self.p:
            return 0

        return 1


class Uniform(Distribution[int]):
    __slots__ = ("a", "b")

    a: Final[int]
    b: Final[int]

    def __init__(self, a: Z, b: Z):
        if not isinstance(a, numbers.Integral):
            raise TypeError("a must be an integral number")
        if not isinstance(b, numbers.Integral):
            raise TypeError("b must be an integral number")
        if b <= a:
            raise ValueError("a must be strictly less then b")

        self.a = int(a)
        self.b = int(b)

    @property
    def n(self) -> int:
        return self.b - self.a + 1

    @property
    def support(self) -> Interval[int]:
        return Interval(self.a, self.b)

    @property
    def mean(self) -> Union[int, Float]:
        if self.n % 2 == 0:
            return (self.a + self.b) // 2
        else:
            return mpmath.fraction(self.a + self.b, 2)

    @property
    def variance(self) -> Float:
        return mpmath.fraction(self.n ** 2 - 1, 12)

    def pmf(self, x: Integral) -> Probability:
        # if x in self.x:
        if x in self.support:
            return mpmath.fraction(1, self.n)

        return mpmath.mpf(0)

    def cmf(self, x: Integral) -> Probability:
        # if x in self.x:
        if self.a <= x <= self.b:
            return (x - self.a + 1) / mpmath.mpf(self.n)
        return mpmath.mp.zero

    def cmf_inf(self, y: Probability) -> int:
        return int(self.n * y + self.a - 1)


# Shortcuts

Bern = Bernoulli
U = Uniform
B = Binomial



from __future__ import annotations


__all__ = [
    "Integral",
    "Real",
    "Complex",
    "Float",
    "Z",
    "R",
    "C",
    "is_integral",
    "is_real",
    "is_complex",
    "is_probability",
    "Interval",
]

import decimal
import fractions
import numbers
import sys
from typing import Any, Generic, Optional, TypeVar, Union, cast


if sys.version_info >= (3, 10):
    from types import EllipsisType  # noqa
else:
    EllipsisType = None

import mpmath
from mpmath import libmp


# numbers.Complex, numbers.Real, etc. as typevar bound are bugged in mypy :(
# https://github.com/python/mypy/issues/3186
Integral = Union[bool, int, libmp.MPZ_TYPE]
Float = Union[float, mpmath.mpf]
Real = Union[Integral, Float, decimal.Decimal, fractions.Fraction]
Complex = Union[Real, complex, mpmath.mpc]


# natural numbers
Z = TypeVar(
    "Z",
    mpmath.libmp.MPZ_TYPE,
    int,
)
# real numbers
R = TypeVar(
    "R",
    mpmath.mpf,
    float,
    decimal.Decimal,
    fractions.Fraction,
    mpmath.libmp.MPZ_TYPE,
    int,
    bool,
)
# complex numbers
C = TypeVar(
    "C",
    mpmath.mpc,
    complex,
    mpmath.mpf,
    float,
    decimal.Decimal,
    fractions.Fraction,
    mpmath.libmp.MPZ_TYPE,
    int,
    bool,
)


def is_integral(x: Any) -> bool:
    return isinstance(x, (int, mpmath.libmp.MPZ_TYPE))


def is_real(x: Any, /, strict: bool = False) -> bool:
    return (
        isinstance(x, (float, decimal.Decimal, fractions.Fraction, mpmath.mpf))
        or not strict
        and is_integral(x)
    )


def is_complex(x: Any, /, strict: bool = False) -> bool:
    return isinstance(x, (complex, mpmath.mpc)) or not strict and is_real(x)


def _cmax(*args: C) -> C:
    res = None
    for arg in args:
        if res is None:
            res = arg
        elif isinstance(res, numbers.Real):  # not complex
            if arg > res:
                res = arg
        else:
            if arg.real > res.real and arg.imag > res.imag:
                res = arg
            elif arg.real > res.real:
                res = type(res)(arg.real, res.imag)
            elif arg.imag > res.imag:
                res = type(res)(res.real, arg.imag)

    if res is None:
        raise ValueError("cmax expected 1 argument, got 0")

    return res


def _cmin(*args: C) -> C:
    return cast(C, -_cmax(*(-arg for arg in args)))


class Interval(Generic[C]):
    """Closed interval (endpoints a and b are included)"""

    a: C
    b: C

    def __init__(self, a: C, b: C, /):
        if a.real > b.real:
            raise ValueError("a cannot be larger than b")
        if a.imag > b.imag:
            raise ValueError("Im(a) cannot be larger than Im(b)")

        self.a = a
        self.b = b

    def __repr__(self):
        return f"{type(self).__name__}({self.a!r}, {self.b!r})"

    def __str__(self):
        return f"[{self.a}, {self.b}]"

    def __contains__(self, x: Union[C, Interval[C], object]):
        if isinstance(x, Interval):
            # subinterval check
            return x.a in self and x.b in self
        if isinstance(x, numbers.Complex):
            return (
                self.a.real <= x.real <= self.b.real
                and self.a.imag <= x.imag <= self.b.imag
            )

        return False

    def __eq__(self, other):
        if isinstance(other, Interval):
            return self.a == other.a and self.b == other.b
        elif self.is_degenerate and isinstance(other, numbers.Complex):
            return other == self.a

        return False

    def __lt__(self, other) -> bool:
        if isinstance(other, Interval):
            # subinterval check
            other = other.a
        if isinstance(other, numbers.Complex):
            return bool(self.b.real < other.real or self.b.imag < other.imag)

        return NotImplemented

    def __le__(self, other) -> bool:
        if isinstance(other, (Interval, numbers.Complex)):
            return self == other or self < other

        return NotImplemented

    def __gt__(self, other) -> bool:
        if isinstance(other, Interval):
            # subinterval check
            other = other.b
        if isinstance(other, numbers.Complex):
            return bool(self.a.real > other.real or self.a.real > other.imag)

        return NotImplemented

    def __ge__(self, other) -> bool:
        if isinstance(other, (Interval, numbers.Complex)):
            return self == other or self > other

        return NotImplemented

    def __and__(self, other: Interval[Any]) -> Optional[Interval[C]]:
        if isinstance(other, Interval):
            if self.isdisjoint(other):
                return None

            return type(self)(_cmax(self.a, other.a), _cmin(self.b, other.b))
        return NotImplemented

    def __or__(self, other: Interval[C]) -> Interval[C]:
        if isinstance(other, Interval):
            if self.isdisjoint(other):
                # TODO mask the gap, or create a disjoint interval class
                raise NotImplementedError

            return type(self)(_cmin(self.a, other.a), _cmax(self.b, other.b))
        return NotImplemented

    def __bool__(self) -> bool:
        return True

    def __hash__(self):
        return hash((self.a, self.b))

    @property
    def is_bounded(self) -> bool:
        return bool(
            mpmath.isfinite(self.a.real)
            and mpmath.isfinite(self.a.imag)
            and mpmath.isfinite(self.b.real)
            and mpmath.isfinite(self.b.imag)
        )

    @property
    def is_degenerate(self) -> bool:
        return bool(self.a == self.b)

    @property
    def size(self) -> Union[C, int]:  # lower type bound for bool
        return self.b - self.a

    @property
    def radius(self) -> Union[C, float]:  # lower type bound for int
        return self.size / 2

    @property
    def mid(self) -> Union[C, float]:  # lower type bound for bool
        return (self.a + self.b) / 2

    def isdisjoint(self, other: Interval[Any]) -> bool:
        return bool(
            self.a.real > other.b.real
            or self.a.imag > other.b.imag
            or self.b.real < other.a.real
            or self.a.imag < other.a.imag
        )


def is_probability(y: Any) -> bool:
    if isinstance(y, numbers.Real):
        return bool(0.0 <= float(y) <= 1.0)

    return False



__all__ = ["MPBackend", "get_backend", "mp_configure"]

from typing import Final, Tuple, Type, TypeVar

import mpmath
from mpmath import ctx_base, libmp
from typing_extensions import Literal


MPBackend = Literal["python", "gmpy", "sage"]

MP_CONTEXTS: Final[Tuple[mpmath.ctx_base.StandardBaseContext, ...]] = (
    mpmath.fp,
    mpmath.mp,
    mpmath.iv,
)


def get_backend() -> MPBackend:
    backend: MPBackend = libmp.BACKEND
    return backend


def mp_configure(
    *,
    prec: int = 53,
    dps: int = 15,
    pretty: bool = True,
) -> None:

    """Configure all mpmath contexts.

    :param prec: binary precision in bits
    :param dps: decimal precision
    :param pretty: pretty formatting for `repr()`
    """
    for context in MP_CONTEXTS:
        context.prec = prec
        context.dps = dps
        context.pretty = pretty



from __future__ import annotations


__all__ = [
    "Stochast",
    "Distribution",
]

import abc
from typing import ClassVar, Generic, Protocol, Union, runtime_checkable

from hall._types import C
from hall._types import Float
from hall._types import Float as Probability
from hall._types import Interval
from hall.event import EventEq, EventInterval


class Stochast(Generic[C]):
    # TODO (linear) arithmatic
    __slots__ = ("distribution",)

    distribution: Distribution[C]

    def __init__(self, distribution: Distribution[C]):
        self.distribution = distribution

    def __eq__(self, other) -> EventEq:  # type: ignore
        return EventEq(self, other)

    def __ne__(self, other) -> EventEq:  # type: ignore
        return EventEq(self, other, _inv=True)

    def __lt__(self, other) -> EventInterval:
        if self.distribution.discrete:
            return EventInterval(self, b=other - 1)
        else:
            return EventInterval(self, b=other)  # TODO some tiny amount

    def __le__(self, other) -> EventInterval:
        return EventInterval(self, b=other)

    def __gt__(self, other) -> EventInterval:
        return EventInterval(self, a=other)

    def __ge__(self, other) -> EventInterval:
        if self.distribution.discrete:
            return EventInterval(self, a=other + 1)
        else:
            return EventInterval(self, a=other)  # TODO some tiny amount

    def __contains__(self, item) -> bool:
        return item in self.distribution.support

    @property
    def outcomes(self) -> Interval[C]:
        return self.distribution.support


@runtime_checkable
class Distribution(Protocol[C]):
    __slots__ = ()

    discrete: ClassVar[bool]

    def __invert__(self) -> Stochast[C]:
        return Stochast(self)

    @property
    @abc.abstractmethod
    def support(self) -> Interval[C]:
        ...

    @property
    @abc.abstractmethod
    def mean(self) -> Union[C, Float]:
        """
        The mean/expected value of the distribution
        """
        ...

    @property
    @abc.abstractmethod
    def variance(self) -> Union[C, Float]:
        """
        The variance of the distribution
        """
        ...

    @abc.abstractmethod
    def f(self, x: C) -> Probability:
        """
        Probability Mass/Density Function (PMF/PDF) that must integrate to 1.
        """
        ...

    @abc.abstractmethod
    def F(self, x: C) -> Probability:
        """
        Cumulative Distribution Function (CDF); the integral of `f`.
        """
        ...

    @abc.abstractmethod
    def G(self, y: Probability) -> C:
        """
        Percent Point Function (PPF); the inverse of the CDF `F`
        """
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hall']

package_data = \
{'': ['*']}

install_requires = \
['mpmath>=1.2,<2.0']

entry_points = \
{'console_scripts': ['docs = scripts.poetry:docs',
                     'reformat = scripts.poetry:reformat',
                     'test = scripts.poetry:test']}

setup_kwargs = {
    'name': 'hall',
    'version': '0.1.4',
    'description': 'Hall - Probability theory using pythonic and (almost) mathematical notation.',
    'long_description': '<!--badges-start-->\n[![CI](https://github.com/jorenham/hall/workflows/CI/badge.svg?event=push)](https://github.com/jorenham/hall/actions?query=event%3Apush+branch%3Amaster+workflow%3ACI)\n[![pypi](https://img.shields.io/pypi/v/hall.svg)](https://pypi.python.org/pypi/hall)\n[![Downloads](https://pepy.tech/badge/hall/month)](https://pepy.tech/project/hall)\n[![versions](https://img.shields.io/pypi/pyversions/hall.svg)](https://github.com/jorenham/hall)\n[![license](https://img.shields.io/github/license/jorenham/hall.svg)](https://github.com/jorenham/hall/blob/master/LICENSE)\n<!--badges-end-->\n\nProbability theory using pythonic and (almost) mathematical notation.\n\n## Help\n\nSee [documentation](https://jorenham.github.io/hall/) for more details.\n\n## A simple example: Intelligence quotient\n\n<!--example-iq-start-->\n```pycon\n>>> from hall import P, E, Std, Normal\n>>> IQ = ~Normal(100, 15)\n>>> E[IQ]\n100.0\n>>> Std[IQ]\n15.0\n>>> P(X >= 130)\n0.0227501319481792\n```\n\nSo the chance of having an IQ (normally distributed with μ=100 and σ=15) of at \nleast 130 is approximately 2.3%.\n<!--example-iq-end-->\n\n## A simple example: Monty ~~Python~~ Hall \n\n`TODO`\n\n## Contributing\n\nFor guidance on setting up a development environment and how to make a\ncontribution to *hall*, see\n[Contributing to Pydantic](https://jorenham.github.io/hall/#contributing).',
    'author': 'Joren Hammudoglu',
    'author_email': 'jhammudoglu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations


__all__ = [
    "Event",
    "EventEq",
    "EventInterval",
]

import abc
from typing import TYPE_CHECKING, Final, Optional, Protocol

import mpmath

from hall._types import C
from hall._types import Float as Probability
from hall._types import is_probability


if TYPE_CHECKING:
    from hall._core import Stochast


class Event(Protocol[C]):
    __slots__ = ()

    X: Stochast[C]

    def __float__(self) -> float:
        """return the probability of the event occurring"""
        return float(self.p)

    def __bool__(self) -> bool:
        """Returns `True` with probability `self.p`, otherwise `False`"""
        return bool(mpmath.rand() < self.p)

    def __int__(self) -> int:
        return int(bool(self))

    @property
    @abc.abstractmethod
    def p(self) -> Probability:
        """return the probability of the event occurring"""
        ...


class EventEq(Event[C]):
    """
    Symbolic (in)equality expression of a stochast and a constant.
    """

    __slots__ = ("X", "x", "_inv")  # noqa

    X: Stochast[C]
    x: C
    _inv: Final[bool]

    def __init__(self, X: Stochast[C], x: C, _inv: bool = False):
        self.X = X
        self.x = x
        self._inv = _inv

    def __repr__(self):
        return " ".join((str(self.X), self._op_symbol, str(self.x)))

    __str__ = __repr__

    def __invert__(self) -> EventEq[C]:
        return type(self)(self.X, self.x, not self._inv)

    def __hash__(self):
        return hash((self.X, self.x, self._inv))

    @property
    def p(self) -> Probability:
        p = self.X.distribution.f(self.x)
        if self._inv:
            return 1.0 - p
        else:
            return p

    @property
    def _op_symbol(self) -> str:
        return "≠" if self._inv else "="


class EventInterval(Event[C]):
    """
    Symbolic "greater/less than" expression of a stochast and a constant.
    """

    __slots__ = ("X", "a", "b", "_inv")  # noqa

    X: Stochast[C]
    a: Optional[C]
    b: Optional[C]
    _inv: Final[bool]

    def __init__(
        self,
        X: Stochast[C],
        a: Optional[C] = None,
        b: Optional[C] = None,
        _inv: bool = False,
    ):
        if a is None and b is None:
            raise ValueError("a or b required")

        self.X = X
        self.a = a
        self.b = b
        self._inv = _inv

    def __repr__(self):
        if self.a is None:
            res = f"{self.X} <= {self.b:.4f}"
        elif self.b is None:
            res = f"{self.X} > {self.a:.4f}"
        else:
            res = f"{self.a:.4f} < {self.X} <= {self.b:.4f}"

        if self._inv:
            return f"~({res})"

        return res

    __str__ = __repr__

    def __invert__(self) -> EventInterval:
        if self.a is None:
            return type(self)(self.X, a=self.b, _inv=self._inv)
        if self.b is None:
            return type(self)(self.X, b=self.a, _inv=self._inv)

        return type(self)(self.X, a=self.a, b=self.b, _inv=not self._inv)

    def __hash__(self):
        return hash((self.X, self.a, self.b, self._inv))

    @property
    def p(self) -> Probability:
        p: Probability
        if self.b is None:
            p = mpmath.mpf(1)
        else:
            p = self.X.distribution.F(self.b)

        if self.a is not None:
            p -= self.X.distribution.F(self.a)

        if self._inv:
            p = mpmath.mpf(1) - p

        assert is_probability(p), f"{p!r} is not a probability"

        return p



__all__ = [
    "Normal",
    "N",
]

import abc
import numbers
import statistics
from typing import ClassVar, Protocol, cast

import mpmath

from hall._core import Distribution as _Distribution
from hall._types import Float as Probability
from hall._types import Interval, R, Real


class Distribution(_Distribution[R], Protocol[R]):
    discrete: ClassVar[bool] = False

    @abc.abstractmethod
    def pdf(self, x: R) -> Probability:
        """Probability Density Function"""
        raise NotImplementedError

    @abc.abstractmethod
    def cdf(self, x: R) -> Probability:
        """Cumulative Density/Distribution Function"""
        raise NotImplementedError

    def cdf_inv(self, y: Probability) -> R:
        """Inverse of the cdf"""
        # TODO
        #  https://mpmath.org/doc/current/calculus/optimization.html#mpmath.calculus.optimization.Ridder
        raise NotImplementedError

    def f(self, x: R) -> Probability:
        if not isinstance(x, numbers.Real):
            raise TypeError("value must be an real number")

        if x not in self.support:
            return mpmath.mpf(0)

        return self.pdf(x)

    def F(self, x: R) -> Probability:
        if not isinstance(x, numbers.Real):
            raise TypeError("value must be an real number")

        support = self.support
        if x < support:
            return mpmath.mpf(0)
        if x > support:
            return mpmath.mpf(1)

        return self.cdf(x)

    def G(self, y: Probability) -> R:
        if not (0 <= y <= 1):
            raise TypeError("probability must be between 0 and 1 inclusively")

        return self.cdf_inv(y)


class Normal(Distribution[mpmath.mpf]):
    __slots__ = ("mu", "sigma")

    mu: mpmath.mpf
    sigma: mpmath.mpf

    def __init__(
        self,
        mu: Real = mpmath.mpf(0),
        sigma: Real = mpmath.mpf(1),
    ):
        if sigma < 0:
            raise ValueError("sigma must be positive")

        self.mu = mpmath.mpf(mu)
        self.sigma = mpmath.mpf(sigma)

    @property
    def support(self):
        return Interval(mpmath.mpf("-inf"), mpmath.mpf("+inf"))

    @property
    def mean(self):
        return self.mu

    @property
    def variance(self):
        return self.sigma * self.sigma

    def pdf(self, x: Real) -> Probability:
        return mpmath.npdf(x, self.mu, self.sigma)

    def cdf(self, x) -> Probability:
        return mpmath.mp.ncdf(x, self.mu, self.sigma)

    def cdf_inv(self, y: Probability) -> R:
        x = statistics._normal_dist_inv_cdf(y, self.mu, self.sigma)  # type: ignore  # noqa
        return cast(R, mpmath.mpf(x))


# shortcuts

N = Normal
